--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_urv_wrapper
--
-- description: A small wrapper for the URV encompassing the internal RAM and
-- access to the RAM through CPU CSR register block.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.genram_pkg.all;
--use work.memory_loader_pkg.all;
use work.wishbone_pkg.all;
use work.urv_pkg.all;
use work.secded_32b_pkg.all;

entity diot_hydra2_core is
  generic(
    g_IRAM_LOG_SIZE : natural := 12;
    g_DRAM_LOG_SIZE : natural := 12);
  port(
    clk_sys_i : in  std_logic;
    clk_cpu_i : in  std_logic_vector(3 downto 1);
    rst_n_i   : in  std_logic;

    --  Set to restart (reset) the cpus.
    restart_i : in  std_logic;

    --  CPU fault (double ecc error, double divergence)
    cpu_flt_o : out std_logic;

    --  CPU self-reset (recovery)
    cpu_rst_o : out std_logic;
    iram_copy_i : in  std_logic;
    dram_clear_i : in  std_logic;
    iram_err_i : in  std_logic;

    dwb_o     : out t_wishbone_master_out;
    dwb_i     : in  t_wishbone_master_in;

    fic0_ready_i   : in  std_logic;

    fic0_haddr_o     : out std_logic_vector(31 downto 0);
    fic0_hburst_o    : out std_logic_vector(2 downto 0);
    fic0_hmastlock_o : out std_logic;
    fic0_hprot_o     : out std_logic_vector(3 downto 0);
    fic0_hsize_o     : out std_logic_vector(2 downto 0);
    fic0_htrans_o    : out std_logic_vector(1 downto 0);
    fic0_hwdata_o    : out std_logic_vector(31 downto 0);
    fic0_hwrite_o    : out std_logic;
    fic0_hrdata_i    : in  std_logic_vector(31 downto 0);
    fic0_hready_i    : in  std_logic;
    fic0_hresp_i     : in  std_logic_vector(1 downto 0);

    led_en_o     : out std_logic;
    leds_o       : out std_logic_vector(4 downto 1)
  );
end diot_hydra2_core;

architecture arch of diot_hydra2_core is

  signal cpu_rst, cpu_rst_n : std_logic;

  signal im_waddr : std_logic_vector(g_IRAM_LOG_SIZE - 1 downto 2);
  signal im_wdata : std_logic_vector(31 downto 0);
  signal im_wecc, im_force_ecc : std_logic_vector(6 downto 0);
  signal im_write : std_logic;
  signal dram_clr : std_logic;
  signal cpu_flt : std_logic;

  --  Last reset was due to a fault
  signal rst_flt : std_logic;
begin

  cpu_rst_n <= not cpu_rst;
  
  inst_hydra: entity work.hydra_core
    generic map (
      g_IRAM_LOG_SIZE => g_IRAM_LOG_SIZE,
      g_DRAM_LOG_SIZE => g_DRAM_LOG_SIZE
    )
    port map (
      clk_sys_i        => clk_sys_i,
      clk_cpu_i        => clk_cpu_i,
      rst_n_i          => rst_n_i,
      cpu_rst_n_i      => cpu_rst_n,
      dwb_o            => dwb_o,
      dwb_i            => dwb_i,
      iram_addr_i      => im_waddr,
      iram_we_i        => im_write,
      iram_data_i      => im_wdata,
      iram_ecc_i       => im_wecc,
      dram_clr_i       => dram_clr,
      cpu_flt_o        => cpu_flt
    );
  
  im_force_ecc <= b"000_0000" when (iram_err_i = '0'
                                    or im_waddr (g_IRAM_LOG_SIZE - 1 downto 4) /= (g_IRAM_LOG_SIZE - 1 downto 4 => '1'))
    else b"000_0001" when im_waddr(2) = '0'
    else b"000_0011";
  im_wecc <= f_calc_ecc (im_wdata) xor im_force_ecc;

  b_init: block
    type t_main_state is (init_delay, init, envm_ready, read_word, wait_word, done);
    signal main_state : t_main_state;

    constant delay : std_logic_vector(27 downto 0) := x"003_0000"
    -- pragma translate off
    xor x"003_0010"
    -- pragma translate_on
    ;

    type t_ahb_state is (idle, addr_phase, wait_rsp);
    signal ahb_state : t_ahb_state;
    signal htrans : std_logic_vector(1 downto 0);

    signal fic_start, fic_done : std_logic;
    signal hpms_data : std_logic_vector (31 downto 0);

    signal envm_addr : std_logic_vector (31 downto 0);
    signal envm_len : unsigned (31 downto 0);

    signal dest_addr : std_logic_vector (31 downto 0);

    signal leds : std_logic_vector(1 downto 0);
  begin
    fic0_hburst_o <= "000";
    fic0_hmastlock_o <= '0';
    fic0_hprot_o <= "0011";
    fic0_hsize_o <= "010";
    fic0_htrans_o <= htrans;
    fic0_haddr_o <= envm_addr;
    fic0_hwrite_o <= '0';

    leds_o(2 downto 1) <= "11";
    leds_o(4 downto 3) <= leds;

    -- Main FSM: copy eNVM to IRAM
    process (clk_sys_i)
    begin
      if rising_edge (clk_sys_i) then
        fic_start <= '0';
        im_write <= '0';
        cpu_rst <= '1';
        led_en_o <= '1';

        if rst_n_i = '0' then
          main_state <= init_delay;
          envm_addr <= x"0000_0000";
          leds <= "11";
          dram_clr <= '0';
          rst_flt <= '0';
        else
          case main_state is
            when init_delay =>
              --  envm_addr is used as a counter
              if envm_addr (31 downto 4) = delay then
                main_state <= init;
              else
                envm_addr <= std_logic_vector(unsigned(envm_addr) + 4);
              end if;
            when init =>
              leds <= "11";
              dram_clr <= dram_clear_i or rst_flt;
              -- poll eNVM ready bit from status register
              envm_addr <= x"6008_0120";
              fic_start <= '1';

              main_state <= envm_ready;

            when envm_ready =>
              leds <= "10";
              if fic_done = '1' then
                if hpms_data (0) = '1' then
                  --  eNVM is ready.
                  if iram_copy_i = '1' or rst_flt = '1' then
                    main_state <= read_word;
                  else
                    main_state <= done;
                  end if;
                
                  --  Copy the whole IRAM
                  envm_addr <= x"6000_0000";
                  envm_len <= to_unsigned(2**g_IRAM_LOG_SIZE, 32);
                  dest_addr <= x"0000_0000";
                else
                  --  eNVM not ready
                  --  poll again.
                  fic_start <= '1';
                  main_state <= envm_ready;
                end if;
              end if;

            when read_word =>
              leds <= "00";
              if fic_done = '0' then
                fic_start <= '1';

                envm_len <= envm_len - 4;

                main_state <= wait_word;
              end if;

            when wait_word =>
              leds <= "00";
              if fic_done = '1' then
                --  Write the word.
                im_wdata <= hpms_data;
                im_write <= '1';
                im_waddr <= dest_addr(g_IRAM_LOG_SIZE - 1 downto 2);

                dest_addr <= std_logic_vector(unsigned(dest_addr) + 4);
                envm_addr <= std_logic_vector(unsigned(envm_addr) + 4);

                if envm_len = 0 then
                  main_state <= done;
                else
                  main_state <= read_word;
                end if;
              end if;
          
            when done =>
              led_en_o <= '0';
              leds <= "00";
              cpu_rst <= '0';
              dram_clr <= '0';

              if restart_i = '1' or cpu_flt = '1' then
                cpu_rst <= '1';
                main_state <= init;
                rst_flt <= cpu_flt;
              end if;
          end case;
        end if;
      end if;
    end process;

    --  AHB FSM
    process (clk_sys_i)
    begin
      if rising_edge (clk_sys_i) then
        htrans <= "00";
        fic_done <= '0';

        if rst_n_i = '0' then
          ahb_state <= idle;
        else
          case ahb_state is
            when idle =>
              if fic_start = '1' then
                htrans <= "10";  --  Non seq
                ahb_state <= addr_phase;
              end if;
            when addr_phase =>
              if fic0_hready_i = '1' then
                ahb_state <= wait_rsp;
              else
                --  Was not ready.
                htrans <= "10";  --  Non seq
              end if;
            when wait_rsp =>
              if fic0_hready_i = '1' then
                ahb_state <= idle;
                fic_done <= '1';
                hpms_data <= fic0_hrdata_i;
              end if;
          end case;
        end if;
      end if;
    end process;
  end block;
end arch;
