#!/usr/bin/env python3

import sys
import time
import subprocess
import random

class FipUrvDrv:
    def __init__(self, trace=False):
        self.proc = subprocess.Popen(["./fip_urv", "-p"],
                                     universal_newlines=True,
                                     bufsize=0,
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE)
        self.trace = trace
        self.nbr_comm_err = 0
        self.nbr_ping_err = 0
        self.nbr_flash_err = 0

    def read_reply(self):
        l = ''
        res = []
        while True:
            c = self.proc.stdout.read(1)
            if c == 0:
                # EOF
                break
            if c == '\n':
                if len(l) > 2 and l[:2] == '# ':
                    break
                res.append(l)
                l = ''
            else:
                l += c
        if self.trace:
            sys.stderr.write("Recv: {} + {}\n".format(res, l))
        if l != "# OK":
            print("ERROR: comm, status: {}".format(l))
            self.nbr_comm_err += 1
        return res

    def send_cmd(self, cmd):
        if self.trace:
            sys.stderr.write("Send: {}\n".format(cmd))
        self.proc.stdin.write(cmd + '\n')


    def cmd(self, cmd):
        self.send_cmd(cmd)
        return self.read_reply()

    def disp(self, res):
        for l in res:
            print(l, end='')

    def close(self):
        self.proc.stdin.close()
        self.read_reply()
        self.proc.wait()


def get_pow_vc(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    iout = s.cmd("mon -s iout")
    return (vout[0], iout[0])

def get_pow_v(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    return vout[0]


def get_cmd(s, cmd):
    res = s.cmd("mon -s {}".format(cmd))
    return res[0]

def show_cmd(s, cmd, name):
    res = get_cmd(s, cmd)
    print("{}:  {}".format(name, res))

REGS = 0x100000
SPI = REGS + 0xe0

def spi_init(s):
    # Set divider (1Mhz)
    s.cmd("write 0x{:x} 20".format(SPI + 0x14))

def spi_cmd(s, cmd):
    """Send cmd to spi.  cmd is a list of bytes"""
    l = len(cmd)
    assert 0 < l <= 8

    # Set ctrl char-len=5*8=0x28, tx_neg=1, rx_neg=0
    s.cmd("write 0x{:x} 0x{:04x}".format(SPI + 0x10, 0x400 + l * 8))

    # Bytes to be sent
    ra = SPI + ((l - 1) // 4) * 4

    w = 0
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        #print("  l={}, i={}, off={}".format(l, i, off))
        w |= cmd[i] << (off * 8)
        if off == 0 or i == l - 1:
            s.cmd("write 0x{:x} 0x{:x}".format(r, w))
            r -= 4
            w = 0

    # Set SS=1
    s.cmd("write 0x{:x} 1".format(SPI + 0x18))

    # Go
    s.cmd("write 0x{:x} 0x{:04x}".format(SPI + 0x10, 0x500 + l * 8))

    # Read
    res = []
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        if i == 0 or off == 3:
            w = s.cmd("read 0x{:x}".format(r))
            r -= 4
            w = int(w[0], 0)
        v = (w >> (off * 8)) & 0xff
        res.append(v)

    # Set SS=1
    s.cmd("write 0x{:x} 0".format(SPI + 0x18))

    return res

def flash_id(s):
    spi_init(s)

    res = spi_cmd(s, [0x05, 0])
    print("Flash status: {:02x}".format(res[1]))

    res = spi_cmd(s, [0x9f, 0, 0, 0, 0])
    print("Flash id: {:02x}{:02x}{:02x}{:02x}".format(
        res[1], res[2], res[3], res[4]))

    # Read
    for i in range(10):
        res = spi_cmd(s, [0x03, 0, i, 0, 0, 0, 0, 0])
        print("Flash 0x{:04x}: {:02x} {:02x} {:02x} {:02x}".format(
            i << 8, res[4], res[5], res[6], res[7]))

def flash_wait_write(s):
    while True:
        res = spi_cmd(s, [0x05, 0])
        print("Flash status: {:02x}".format(res[1]))
        if (res[1] & 0x20) != 0:
            print("ERROR: flash write/erase")
            s.nbr_flash_err += 1
            break
        if (res[1] & 1) == 0:
            break

def flash_prg(s, pg, off):
    print("Write flash at page={:02x} off={:02x}".format(pg, off))

    # Write enable
    spi_cmd(s, [0x06])

    # Write
    npg = (pg + 1) & 0xff
    spi_cmd(s, [0x02, 0x00, pg, off, pg, (~pg) & 0xff, npg, (~pg) & 0xff])

    flash_wait_write(s)

    # Write disable
    spi_cmd(s, [0x04])


def flash_full_erase(s):
    print("Erasing flash")

    # Write enable
    spi_cmd(s, [0x06])

    # Erase
    spi_cmd(s, [0x60])

    flash_wait_write(s)

def flash_page_erase(s, pg):
    print("Erasing flash page 0x{:02x}".format(pg))

    # Write enable
    spi_cmd(s, [0x06])

    # Page erase
    spi_cmd(s, [0x81, 0x00, pg, 0x00])

    flash_wait_write(s)

def flash_find_free(s):
    """Assuming flash pages are written consecutively, find the first erased
    one"""
    first = 0
    last = 255

    print("Find first free page:")
    while True:
        mid = (first + last) // 2
        res = spi_cmd(s, [0x03, 0, mid, 0, 0, 0])
        print("Page {}: first byte is {:02x} {:02x} (first:{}, last:{})".format(
            mid, res[4], res[5], first, last))
        if res[4] == 0xff and res[5] == 0xff:
            # Page is free
            if mid == first:
                return mid
            last = mid - 1
        else:
            # Page is busy
            if mid == last:
                if last == 255:
                    return -1
                return last + 1
            first = mid + 1

def show_psu(s):
    psu = s.cmd("psu")
    s.cmd("i2c 2")
    print("********** PSU 0 ******************")
    if psu[0] == "PSU 0: presence: 0":
        print("Not present")
    elif psu[0] == "PSU 0: presence: 1":
        print("TODO")
    else:
        print("unhandled")
    print("********** PSU 1 ******************")
    if psu[1] == "PSU 1: presence: 0":
        print("Not present")
    elif psu[1] == "PSU 1: presence: 1":
        show_cmd(s, "-t 0x18 vout", "Vout")
        show_cmd(s, "-t 0x18 iout", "Iout")
        show_cmd(s, "-t 0x18 temp-1", "Temp 1")
        show_cmd(s, "-t 0x18 temp-2", "Temp 2")
    else:
        print("unhandled")

def show_boards(s):
    print("********** Configuration *********")
    s.cmd("i2c 1")
    print("SLOTS: 01 02 03 04 05 06 07 08")
    print("      ", end='')
    for i in range(8):
        res = s.cmd("i2c-id -s {}".format(i))
        print(" {}".format(res[0]), end='')
    print()

def show_sysmon(s):
    s.cmd("i2c 0")
    t = time.time()
    ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    rev = get_cmd(s, "revision")
    print("******** System board (revision {}) ********".format(rev))
    print("time: {}".format(ts))
    p12 = get_pow_vc(s, 0)
    print("P12V:   {}   {}".format(p12[0], p12[1]))
    p1_2 = get_pow_v(s, 1)
    print("P1.2V:  {}".format(p1_2))
    p3_3 = get_pow_v(s, 2)
    print("P3.3V:  {}".format(p3_3))
    pperi = get_pow_v(s, 3)
    print("Vperi:  {}".format(pperi))
    tamb = get_cmd(s, "temp-1")
    print("Temp amb:  {}".format(tamb))
    tldo = get_cmd(s, "temp-2")
    print("Temp LDO:  {}".format(tldo))
    tfpga = get_cmd(s, "temp-3")
    print("Temp FPGA: {}".format(tfpga))
    print("CSV: {tm},{p12v},{p12i},{p1_2},{p3_3},{tamb},{tldo},{tfpga},{ncom_err},{nflash_err},{nping_err}".format(
        tm=ts,
        p12v=p12[0].rstrip(' V'),
        p12i=p12[1].rstrip(' A'),
        p1_2=p1_2.rstrip(' V'),
        p3_3=p3_3.rstrip(' V'),
        tamb=tamb.rstrip(' C'),
        tldo=tldo.rstrip(' C'),
        tfpga=tfpga.rstrip(' C'),
        ncom_err=s.nbr_comm_err,
        nflash_err=s.nbr_flash_err,
        nping_err=s.nbr_ping_err))

rnd = random.Random()

def do_ping(s):
    v = rnd.randint(0, (1<<32) - 1)
    print("Ping with value {:08x}".format(v))
    s.cmd("write 0x107fc 0x{:08x}".format(v))
    w = s.cmd("read 0x107fc")
    rv = int(w[0], 0)
    if rv != v:
        print("ERROR: ping (write {:08x}, read {:08x})".format(v, rv))
        s.nbr_ping_err += 1

def extract(s, left, right=None):
    if len(s) < len(left):
        return "ERR1"
    if s[:len(left)] != left:
        return "ERR2"
    s = s[len(left):]
    if right is None:
        return s
    if len(s) < len(right):
        return "ERR3"
    if s[-len(right):] != right:
        return "ERR4"
    return s[:-len(right)]

def do_extract_recovery(s):
    p = s.cmd("stat-regs")
    nrecv = int(extract(p[2], "nbr recovery: "), 0)
    nrst = int (extract(p[3], "nbr resets: "), 0)
    last = int(extract(p[4], "last reset: "), 0)
    return (nrecv, nrst, last)

def do_self_tests_recovery(s):
    print("* Test CPU recovery")
    nrecv, nrst, _ = do_extract_recovery(s)
    for i in range(3):
        print("Fault on cpu {}".format(i))
        s.cmd("cpu-fault {}".format(i))
        nrecv1, nrst1, last = do_extract_recovery(s)
        if nrecv1 != nrecv + 1:
            print("Bad nbr recovery: read {}, expect {}".format(
                nrecv1, nrecv + 1))
        else:
            nrecv += 1
        if nrst1 != nrst + 1:
            print("Bad nbr reset: read {}, expect {}".format(
                nrst1, nrst + 1))
        else:
            nrst += 1
        if last != (1 << i):
            print("Bad last reset: read {}, expect {}".format(last, 1 << i))
            cause = do_extract_reset_cause(s)
            if cause != 1:
                print("bad reset cause, read {}, expect 1".format(cause))

def do_extract_reset_cause(s):
    p = s.cmd("supervisor")
    return int(extract(p[0], "reset cause: "), 0)

def do_extract_dram_ecc(s):
    p = s.cmd("supervisor")
    n_ecc_se = int(extract(p[11], "DRAM corrected errors by scrubber: "), 0)
    n_ecc_de = int(extract(p[13],
                           "DRAM double errors detected in last cycle: "), 0)
    return (n_ecc_se, n_ecc_de)

def do_self_tests_dram(s):
    print("* Test DRAM ecc")
    n_ecc_se, n_ecc_de = do_extract_dram_ecc(s)
    print("se: {}, de: {}".format(n_ecc_se, n_ecc_de))
    if n_ecc_de != 0:
        print("Already data ECC errors")
    s.cmd("write-ecc 0x40100 0x123 1")
    n_ecc_se1, n_ecc_de1 = do_extract_dram_ecc(s)
    print("se: {}, de: {}".format(n_ecc_se1, n_ecc_de1))
    if n_ecc_se1 != n_ecc_se + 1 or n_ecc_de1 != 0:
        print("bad number of ECC errors")
    n_ecc_se = n_ecc_se1
    s.cmd("write-ecc 0x40104 0x456 6")
    n_ecc_se1, n_ecc_de1 = do_extract_dram_ecc(s)
    print("se: {}, de: {}".format(n_ecc_se1, n_ecc_de1))
    if n_ecc_se1 != n_ecc_se or n_ecc_de1 != 1:
        print("bad number of ECC errors")
    s.cmd("read 0x40104")
    cause = do_extract_reset_cause(s)
    if cause != 4:
        print("bad reset cause, read {}, expect 4".format(cause))
    s.cmd("write-ecc 0x40104 0 0")

def do_extract_iram_ecc(s):
    p = s.cmd("supervisor")
    n_ecc_se = int(extract(p[6], "IRAM corrected errors by scrubber: "), 0)
    n_ecc_de = int(extract(p[8],
                           "IRAM double errors detected in last cycle: "), 0)
    return (n_ecc_se, n_ecc_de)

def do_self_tests_iram(s):
    print("* Test IRAM ecc")
    n_ecc_se, n_ecc_de = do_extract_iram_ecc(s)
    print("se: {}, de: {}".format(n_ecc_se, n_ecc_de))
    if n_ecc_de != 0:
        print("Already insn ECC errors")
    s.cmd("restart-cfg 7")
    print("crash...")
    s.cmd("write-ecc 0x40104 0x456 6")
    s.cmd("read 0x40104")
    # Will timeout
    cause = do_extract_reset_cause(s)
    if cause != 4:
        print("bad reset cause, read {}, expect 4".format(cause))
    s.cmd("restart-cfg 3")
    n_ecc_se1, n_ecc_de1 = do_extract_iram_ecc(s)
    print("se: {}, de: {}".format(n_ecc_se1, n_ecc_de1))
    if n_ecc_se1 != n_ecc_se + 2 or n_ecc_de1 != 2:
        print("Error in IRAM ecc")
    # Crash too
    s.cmd("read 0xfffc")
    cause = do_extract_reset_cause(s)
    if cause != 2:
        print("bad reset cause, read {}, expect 2".format(cause))

def do_extract_monimod_cycle(s):
    p = s.cmd("monimod-read")
    return int(extract(p[0], "Cycle: "), 0)

def do_extract_monimod_reset(s):
    p = s.cmd("stat-regs")
    return int(extract(p[0], "nbr monimod reset: "), 0)

def do_self_tests_monimod_scan(s, cnt=20):
    cyc = do_extract_monimod_cycle(s)
    ntry = cnt
    while ntry > 0:
        ncyc = do_extract_monimod_cycle(s)
        if ncyc != cyc:
            break
        ntry -= 1
    if ntry == 0:
        print("no monimod scan")

def do_self_tests_monimod(s):
    print('* Test monimod')
    do_self_tests_monimod_scan(s)
    nrst = do_extract_monimod_reset(s)
    s.cmd("monimod-power off")
    do_self_tests_monimod_scan(s, 100)
    nrst1 = do_extract_monimod_reset(s)
    if nrst1 != nrst + 1:
        print("monimod reset expected")

def do_self_tests(s):
    do_self_tests_recovery(s)
    do_self_tests_dram(s)
    do_self_tests_iram(s)
#    do_self_tests_monimod(s)

def run():
    s = FipUrvDrv(trace=False)
    # s.cmd("slots")

    verbose = False
    flag_dry = False
    flag_flash = True

    for arg in sys.argv[1:]:
        if arg == '-v':
            verbose = True
        elif arg == '--dry' or arg == '-n':
            flag_dry = True
        elif arg == '--no-flash':
            flag_flash = False
        elif arg == '-t':
            s.trace = True
        else:
            print("unknown option: {}".format(arg))
            sys.exit(1)

    s.cmd("monimod-scan on")
    s.cmd("plc reset")
    s.cmd("plc halt")
    s.cmd("load plc_mbox.elf")

    if flag_flash:
        flash_id(s)

    print("Self-tests sequence")
    do_self_tests(s)

    if flag_dry:
        return

    flag_erase = False

    # Find first free page
    if flag_flash:
        pg = flash_find_free(s)
        if pg < 0:
            print("No free page")
            flash_full_erase(s)
            pg = 0
            print("First free page: {}".format(pg))
        off = 0

    mbx_exp = 0x12345678

    ping_err = 0
    mbx_err = 0

    first = True
    while True:
        t = time.time()
        ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
        csv = [('time', ts)]
        p = s.cmd("ping")
        if verbose:
            print("Ping: {}".format(p))
        if p[0].find('!') != -1:
            ping_err += 1
        csv.append(('ping', str(ping_err)))
        p = s.cmd("fip-err")
        if verbose:
            print("fip-err: {}".format(p))
        csv.append(('fip-err',
                    extract(p[0], "fip errors: ")))
        csv.append(('comm-err', str(s.nbr_comm_err)))
        p = s.cmd("monimod-read")
        if verbose:
            print("Monimod: {}".format(p))
        csv.append(('cycle', extract(p[0], "Cycle: ")))
        csv.append(('P12V', extract(p[1], "P12V: ", " V")))
        csv.append(('P12A', extract(p[2], "P12V: ", " A")))
        csv.append(('P1.2V', extract(p[3], "P1.2V: ", " V")))
        csv.append(('P3.3V', extract(p[4], "P3.3V: ", " V")))
        csv.append(('Vperi', extract(p[5], "Vperi: ", " V")))
        csv.append(('Tamb', extract(p[6], "Tamb: ", " C")))
        csv.append(('Tldo', extract(p[7], "Tldo: ", " C")))
        csv.append(('Tfpga', extract(p[8], "Tfpga: ", " C")))
        p = s.cmd("supervisor")
        if verbose:
            print("supervisor: {}".format(p))
        csv.append(('reset-cause', extract(p[0], "reset cause: ")))
        csv.append(('cpus-reset', extract(p[1], "cpus on reset: ")))
        csv.append(('iram-se-corr', extract(p[4], "IRAM corrected errors: ")))
        csv.append(('iram-scrub-cyc',
                    extract(p[5], "IRAM scrub cycle number: ")))
        csv.append(('iram-scrub-se',
                    extract(p[6], "IRAM corrected errors by scrubber: ")))
        csv.append(('iram-scrub-de',
                    extract(p[8], "IRAM double errors detected in last cycle: ")))
        csv.append(('dram-se-corr',
                    extract(p[9], "DRAM corrected errors: ")))
        csv.append(('dram-scrub-cyc',
                    extract(p[10], "DRAM scrub cycle number: ")))
        csv.append(('dram-scrub-se',
                    extract(p[11], "DRAM corrected errors by scrubber: ")))
        csv.append(('dram-scrub-de',
                    extract(p[13], "DRAM double errors detected in last cycle: ")))
        csv.append(('insn-bus-err',
                    extract(p[14], "Errors on cpu data bus: ")))
        csv.append(('data-bus-err',
                    extract(p[15], "Errors on cpu insn bus: ")))
        csv.append(('ram-scrub-cfg',
                    extract(p[16], "Scrubber config: ")))
        p = s.cmd("stat-regs")
        if verbose:
            print("stat-regs: {}".format(p))
        csv.append(('monimod-reset',
                    extract(p[0], "nbr monimod reset: ")))
        csv.append(('min-watchdog',
                    extract(p[1], "minimum watchdog count: ")))
        csv.append(('nbr-recovery',
                    extract(p[2], "nbr recovery: ")))
        csv.append(('nbr-reset',
                    extract(p[3], "nbr resets: ")))
        csv.append(('last-reset',
                    extract(p[4], "last reset: ")))
        csv.append(('cycle-count',
                    extract(p[5], "cycle count: ")))
        csv.append(('fip-cmds',
                    extract(p[6], "fip commands: ")))
        csv.append(('fip-pad',
                    extract(p[7], "fip pad: ")))
        p = s.cmd("plc")
        if verbose:
            print("plc: {}".format(p))
        csv.append(('plc-reset',
                    extract(p[0], "plc reset: ")))
        csv.append(('plc-status',
                    extract(p[1], "plc status: ")))
        p = s.cmd("plc-supervisor")
        if verbose:
            print("plc-supervisor: {}".format(p))
        csv.append(('plc-reset-cause',
                    extract(p[0], "PLC reset cause: ")))
        csv.append(('plc-iram-se-corr',
                    extract(p[1], "PLC IRAM corrected errors: ")))
        csv.append(('plc-iram-scrub-cyc',
                    extract(p[2], "PLC IRAM scrub cycle number: ")))
        csv.append(('plc-iram-scrub-se',
                    extract(p[3], "PLC IRAM corrected errors by scrubber: ")))
        csv.append(('plc-iram-scrub-de',
                    extract(p[4], "PLC IRAM double errors detected in last cycle: ")))
        csv.append(('plc-dram-se-corr',
                    extract(p[5], "PLC DRAM corrected errors: ")))
        csv.append(('plc-dram-scrub-cyc',
                    extract(p[6], "PLC DRAM scrub cycle number: ")))
        csv.append(('plc-dram-scrub-se',
                    extract(p[7], "PLC DRAM corrected errors by scrubber: ")))
        csv.append(('plc-dram-scrub-de',
                    extract(p[8], "PLC DRAM double errors detected in last cycle: ")))
        csv.append(('plc-scrub-cfg',
                    extract(p[9], "PLC scrub config: ")))
        p = s.cmd("mbox")
        if verbose:
            print("mbox: {}".format(p))
        mbx_sta = extract(p[0], "mailbox status: ")
        mbx_val = int(extract(p[1], "mailbox in value: "), 0)
        if mbx_sta != '1':
            print("mailbox not ready")
            mbx_err += 1
        else:
            if mbx_val != mbx_exp:
                print("mailbox error: expect {:08x}, read {:08x}".format(
                    mbx_exp, mbx_val))
                mbx_err += 1
            mbx_nval = rnd.randint(0, (1<<32) - 1)
            s.cmd("mbox 0x{:08x}".format(mbx_nval))
            mbx_exp = (~mbx_nval) & 0xffffffff

        csv.append(('mbx-err', str(mbx_err)))

        csv.append(('flash-err', str(s.nbr_flash_err)))

        if first:
            print(','.join(t[0] for t in csv))
            first = False
        print(','.join(t[1] for t in csv))

        sys.stdout.flush()

        if flag_flash:
            if flag_erase:
                flash_page_erase(s, pg)
            flash_prg(s, pg, off)
            pg = (pg + 1) & 0xff
            if pg == 0:
                off = (off + 32) & 0xff
                flag_erase = (off == 0)

    s.close()

run()
