#!/usr/bin/env python3

import sys
import time
import subprocess
import random

class FipUrvDrv:
    def __init__(self, trace=False):
        self.proc = subprocess.Popen(["./fip_urv", "-p"],
                                     universal_newlines=True,
                                     bufsize=0,
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE)
        self.trace = trace
        self.nbr_comm_err = 0
        self.nbr_ping_err = 0
        self.nbr_flash_err = 0

    def read_reply(self):
        l = ''
        res = []
        while True:
            c = self.proc.stdout.read(1)
            if not c:
                break
            if c == '\n':
                if len(l) > 2 and l[:2] == '# ':
                    break
                res.append(l)
                l = ''
            else:
                l += c
        if l != "# OK":
            print("ERROR: comm, status: {}".format(l))
            self.nbr_comm_err += 1
        return res

    def send_cmd(self, cmd):
        if self.trace:
            sys.stderr.write("Send: {}\n".format(cmd))
        self.proc.stdin.write(cmd + '\n')


    def cmd(self, cmd):
        self.send_cmd(cmd)
        return self.read_reply()

    def disp(self, res):
        for l in res:
            print(l, end='')

    def close(self):
        self.proc.stdin.close()
        self.read_reply()
        self.proc.wait()


def get_pow_vc(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    iout = s.cmd("mon -s iout")
    return (vout[0], iout[0])

def get_pow_v(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    return vout[0]


def get_cmd(s, cmd):
    res = s.cmd("mon -s {}".format(cmd))
    return res[0]

def show_cmd(s, cmd, name):
    res = get_cmd(s, cmd)
    print("{}:  {}".format(name, res))

def spi_init(s):
    # Set divider (1Mhz)
    s.cmd("write 0x100094 20")

def spi_cmd(s, cmd):
    """Send cmd to spi.  cmd is a list of bytes"""
    l = len(cmd)
    assert 0 < l <= 8

    # Set ctrl char-len=5*8=0x28, tx_neg=1, rx_neg=0
    s.cmd("write 0x100090 0x{:04x}".format(0x400 + l * 8))

    # Bytes to be sent
    ra = 0x100080 + ((l - 1) // 4) * 4

    w = 0
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        #print("  l={}, i={}, off={}".format(l, i, off))
        w |= cmd[i] << (off * 8)
        if off == 0 or i == l - 1:
            s.cmd("write 0x{:x} 0x{:x}".format(r, w))
            r -= 4
            w = 0

    # Set SS=1
    s.cmd("write 0x100098 1")

    # Go
    s.cmd("write 0x100090 0x{:04x}".format(0x500 + l * 8))

    # Read
    res = []
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        if i == 0 or off == 3:
            w = s.cmd("read 0x{:x}".format(r))
            r -= 4
            w = int(w[0], 0)
        v = (w >> (off * 8)) & 0xff
        res.append(v)

    # Set SS=1
    s.cmd("write 0x100098 0")

    return res

def flash_id(s):
    spi_init(s)

    res = spi_cmd(s, [0x05, 0])
    print("Flash status: {:02x}".format(res[1]))

    res = spi_cmd(s, [0x9f, 0, 0, 0, 0])
    print("Flash id: {:02x}{:02x}{:02x}{:02x}".format(
        res[1], res[2], res[3], res[4]))

    # Read
    for i in range(10):
        res = spi_cmd(s, [0x03, 0, i, 0, 0, 0, 0, 0])
        print("Flash 0x{:04x}: {:02x} {:02x} {:02x} {:02x}".format(
            i << 8, res[4], res[5], res[6], res[7]))

def flash_wait_write(s):
    while True:
        res = spi_cmd(s, [0x05, 0])
        print("Flash status: {:02x}".format(res[1]))
        if (res[1] & 0x20) != 0:
            print("ERROR: flash write/erase")
            s.nbr_flash_err += 1
            break
        if (res[1] & 1) == 0:
            break

def flash_prg(s, pg, off):
    print("Write flash at page={:02x} off={:02x}".format(pg, off))

    # Write enable
    spi_cmd(s, [0x06])

    # Write
    npg = (pg + 1) & 0xff
    spi_cmd(s, [0x02, 0x00, pg, off, pg, (~pg) & 0xff, npg, (~pg) & 0xff])

    flash_wait_write(s)

    # Write disable
    spi_cmd(s, [0x04])


def flash_full_erase(s):
    print("Erasing flash")

    # Write enable
    spi_cmd(s, [0x06])

    # Erase
    spi_cmd(s, [0x60])

    flash_wait_write(s)

def flash_page_erase(s, pg):
    print("Erasing flash page 0x{:02x}".format(pg))

    # Write enable
    spi_cmd(s, [0x06])

    # Page erase
    spi_cmd(s, [0x81, 0x00, pg, 0x00])

    flash_wait_write(s)

def flash_find_free(s):
    """Assuming flash pages are written consecutively, find the first erased
    one"""
    first = 0
    last = 255

    print("Find first free page:")
    while True:
        mid = (first + last) // 2
        res = spi_cmd(s, [0x03, 0, mid, 0, 0, 0])
        print("Page {}: first byte is {:02x} {:02x} (first:{}, last:{})".format(
            mid, res[4], res[5], first, last))
        if res[4] == 0xff and res[5] == 0xff:
            # Page is free
            if mid == first:
                return mid
            last = mid - 1
        else:
            # Page is busy
            if mid == last:
                if last == 255:
                    return -1
                return last + 1
            first = mid + 1

def show_psu(s):
    psu = s.cmd("psu")
    s.cmd("i2c 2")
    print("********** PSU 0 ******************")
    if psu[0] == "PSU 0: presence: 0":
        print("Not present")
    elif psu[0] == "PSU 0: presence: 1":
        print("TODO")
    else:
        print("unhandled")
    print("********** PSU 1 ******************")
    if psu[1] == "PSU 1: presence: 0":
        print("Not present")
    elif psu[1] == "PSU 1: presence: 1":
        show_cmd(s, "-t 0x18 vout", "Vout")
        show_cmd(s, "-t 0x18 iout", "Iout")
        show_cmd(s, "-t 0x18 temp-1", "Temp 1")
        show_cmd(s, "-t 0x18 temp-2", "Temp 2")
    else:
        print("unhandled")

def show_boards(s):
    print("********** Configuration *********")
    s.cmd("i2c 1")
    print("SLOTS: 01 02 03 04 05 06 07 08")
    print("      ", end='')
    for i in range(8):
        res = s.cmd("i2c-id -s {}".format(i))
        print(" {}".format(res[0]), end='')
    print()

def show_sysmon(s):
    s.cmd("i2c 0")
    t = time.time()
    ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    rev = get_cmd(s, "revision")
    print("******** System board (revision {}) ********".format(rev))
    print("time: {}".format(ts))
    p12 = get_pow_vc(s, 0)
    print("P12V:   {}   {}".format(p12[0], p12[1]))
    p1_2 = get_pow_v(s, 1)
    print("P1.2V:  {}".format(p1_2))
    p3_3 = get_pow_v(s, 2)
    print("P3.3V:  {}".format(p3_3))
    pperi = get_pow_v(s, 3)
    print("Vperi:  {}".format(pperi))
    tamb = get_cmd(s, "temp-1")
    print("Temp amb:  {}".format(tamb))
    tldo = get_cmd(s, "temp-2")
    print("Temp LDO:  {}".format(tldo))
    tfpga = get_cmd(s, "temp-3")
    print("Temp FPGA: {}".format(tfpga))
    print("CSV: {tm},{p12v},{p12i},{p1_2},{p3_3},{tamb},{tldo},{tfpga},{ncom_err},{nflash_err},{nping_err}".format(
        tm=ts,
        p12v=p12[0].rstrip(' V'),
        p12i=p12[1].rstrip(' A'),
        p1_2=p1_2.rstrip(' V'),
        p3_3=p3_3.rstrip(' V'),
        tamb=tamb.rstrip(' C'),
        tldo=tldo.rstrip(' C'),
        tfpga=tfpga.rstrip(' C'),
        ncom_err=s.nbr_comm_err,
        nflash_err=s.nbr_flash_err,
        nping_err=s.nbr_ping_err))

rnd = random.Random()

def do_ping(s):
    v = rnd.randint(0, (1<<32) - 1)
    print("Ping with value {:08x}".format(v))
    s.cmd("write 0x107fc 0x{:08x}".format(v))
    w = s.cmd("read 0x107fc")
    rv = int(w[0], 0)
    if rv != v:
        print("ERROR: ping (write {:08x}, read {:08x})".format(v, rv))
        s.nbr_ping_err += 1

def run():
    s = FipUrvDrv(trace=False)
    # s.cmd("slots")

    show_sysmon(s)
    flash_id(s)

    flag_erase = False

    # Find first free page
    pg = flash_find_free(s)
    if pg < 0:
        print("No free page")
        flash_full_erase(s)
        pg = 0
    print("First free page: {}".format(pg))
    off = 0

    while True:
        show_sysmon(s)
        do_ping(s)
        if flag_erase:
            flash_page_erase(s, pg)
        flash_prg(s, pg, off)
        sys.stdout.flush()
        pg = (pg + 1) & 0xff
        if pg == 0:
            off = (off + 32) & 0xff
            flag_erase = (off == 0)

    s.close()

run()
